from __future__ import annotations

from resonate.cmd_queue import CommandQ, Invoke
from resonate.delay_queue import DelayQueue


def test_delay_queue() -> None:
    queue = DelayQueue()
    cmd_queue = CommandQ()
    queue.start(cmd_queue, "")
    queue.enqueue(item=Invoke("3"), delay=0.002)
    assert cmd_queue.get() == Invoke("3")
    cmd_queue.task_done()
    queue.enqueue(item=Invoke("10"), delay=0.01)
    queue.enqueue(item=Invoke("2"), delay=0.002)
    assert cmd_queue.get() == Invoke("2")
    cmd_queue.task_done()
    assert cmd_queue.get() == Invoke("10")
    cmd_queue.task_done()
    queue.stop()



"""New release."""

from __future__ import annotations

import pathlib
import webbrowser
from typing import Any
from urllib.parse import urlencode

import tomli


def project_project(cwd: pathlib.Path) -> dict[str, Any]:
    """Read `pyproject.toml` project."""
    pyproject = tomli.loads(
        cwd.joinpath(
            "pyproject.toml",
        ).read_text(encoding="utf-8"),
    )

    return pyproject["project"]


def main() -> None:
    """Prepare new release."""
    cwd = pathlib.Path().cwd()
    project = project_project(cwd=cwd)
    version = project["version"]
    source = project["urls"]["Source"]
    params = urlencode(
        query={
            "title": f"v{version}",
            "tag": f"v{version}",
        },
    )
    webbrowser.open_new_tab(url=f"{source}/releases/new?{params}")


if __name__ == "__main__":
    main()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from resonate.traits import Subsystem

if TYPE_CHECKING:
    from resonate.dataclasses import SQE


class IProcessor(Subsystem, ABC):
    @abstractmethod
    def enqueue(self, sqe: SQE[Any]) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from resonate.options import Options


class FunctionRegistry:
    def __init__(self) -> None:
        self._store: dict[str, tuple[Callable[[Any], Any], Options]] = {}
        self._index: dict[Callable[[Any], Any], str] = {}

    def add(self, key: str, value: tuple[Callable[[Any], Any], Options]) -> None:
        assert key not in self._store
        assert value[0] not in self._index
        self._store[key] = value
        self._index[value[0]] = key

    def get(self, key: str) -> tuple[Callable[[Any], Any], Options] | None:
        return self._store.get(key)

    def get_from_value(self, v: Callable[[Any], Any]) -> str | None:
        return self._index.get(v)



from __future__ import annotations

from concurrent.futures import Future
from dataclasses import dataclass, field
from typing import Generic, TypeVar, final

from typing_extensions import assert_never

from resonate.result import Err, Ok, Result

T = TypeVar("T")


@final
@dataclass(frozen=True)
class Handle(Generic[T]):
    id: str
    f: Future[T] = field(repr=False, default_factory=Future, init=False)

    def result(self, timeout: float | None = None) -> T:
        return self.f.result(timeout=timeout)

    def set_result(self, result: Result[T, Exception]) -> None:
        if isinstance(result, Ok):
            self.f.set_result(result.unwrap())
        elif isinstance(result, Err):
            self.f.set_exception(result.err())
        else:
            assert_never(result)



from __future__ import annotations

from asyncio import iscoroutinefunction
from inspect import isfunction, isgeneratorfunction
from typing import TYPE_CHECKING, Any, Generic, TypeVar, final

from resonate.actions import LFI, RFI
from resonate.dataclasses import Invocation
from resonate.logging import logger
from resonate.result import Ok, Result
from resonate.retry_policy import Never, exponential, never
from resonate.stores.record import DurablePromiseRecord, TaskRecord

if TYPE_CHECKING:
    from resonate import retry_policy
    from resonate.actions import LFI
    from resonate.context import Context
    from resonate.dataclasses import ResonateCoro
    from resonate.stores.record import DurablePromiseRecord, TaskRecord

T = TypeVar("T")


@final
class Record(Generic[T]):
    def __init__(
        self,
        id: str,
        invocation: LFI | RFI,
        parent: Record[Any] | None,
        ctx: Context,
    ) -> None:
        self.id: str = id
        self.parent: Record[Any] | None = parent
        self.is_root: bool = (
            True if self.parent is None else isinstance(invocation, RFI)
        )
        self._result: Result[T, Exception] | None = None
        self.children: list[Record[Any]] = []
        self.invocation: LFI | RFI = invocation
        self.retry_policy: retry_policy.RetryPolicy | None
        if not isinstance(invocation.unit, Invocation):
            self.retry_policy = None
        elif isinstance(invocation.unit.fn, str):
            self.retry_policy = invocation.opts.retry_policy or None
        elif isgeneratorfunction(invocation.unit.fn):
            self.retry_policy = invocation.opts.retry_policy or never()
        else:
            assert iscoroutinefunction(invocation.unit.fn) or isfunction(
                invocation.unit.fn
            )
            self.retry_policy = invocation.opts.retry_policy or exponential(
                base_delay=1,
                factor=2,
                max_retries=-1,
                max_delay=30,
            )

        self._attempt: int = 1
        self.durable_promise: DurablePromiseRecord | None = None
        self._task: TaskRecord | None = None
        self.ctx = ctx
        self._coro: ResonateCoro[T] | None = None
        self.blocked_on: Record[Any] | None = None
        self._num_children: int = 0
        logger.info(
            "New record %s %s created. Child of %s",
            type(self.invocation).__name__,
            self.id,
            self.parent.id if self.parent else None,
        )

    def get_coro(self) -> ResonateCoro[T]:
        assert self._coro
        return self._coro

    def has_coro(self) -> bool:
        return self._coro is not None

    def increate_attempt(self) -> None:
        self._attempt += 1
        if self._coro is not None:
            self._coro = None

    def root(self) -> Record[Any]:
        maybe_is_the_root = self
        while True:
            if maybe_is_the_root.is_root:
                return maybe_is_the_root
            assert maybe_is_the_root.parent is not None
            maybe_is_the_root = maybe_is_the_root.parent

    def add_child(self, record: Record[Any]) -> None:
        self.children.append(record)
        self._num_children += 1

    def add_coro(self, coro: ResonateCoro[T]) -> None:
        assert self._coro is None
        self._coro = coro

    def add_durable_promise(self, durable_promise: DurablePromiseRecord) -> None:
        assert self.id == durable_promise.id
        assert self.durable_promise is None
        self.durable_promise = durable_promise
        logger.info("Durable promise added to %s", self.id)

    def add_task(self, task: TaskRecord) -> None:
        assert not self.has_task()
        self._task = task
        logger.info("Task added to %s", self.id)

    def has_task(self) -> bool:
        return self._task is not None

    def remove_task(self) -> None:
        assert self.has_task()
        self._task = None
        logger.info("Task completed for %s", self.id)

    def get_task(self) -> TaskRecord:
        assert self._task
        return self._task

    def clear_coro(self) -> None:
        assert self._coro is not None
        self._coro = None

    def should_retry(self, result: Result[T, Exception]) -> bool:
        assert self.retry_policy is not None
        if isinstance(result, Ok) or isinstance(self.retry_policy, Never):
            return False
        return self.retry_policy.should_retry(self._attempt)

    def next_retry_delay(self) -> float:
        assert self.retry_policy is not None
        assert not isinstance(self.retry_policy, Never)
        return self.retry_policy.calculate_delay(self._attempt)

    def set_result(self, result: Result[T, Exception], *, deduping: bool) -> None:
        if not deduping:
            assert not self.should_retry(
                result
            ), "Result cannot be set if retry policy allows for another attempt."

        assert all(
            r.done() for r in self.children
        ), "All children record must be completed."
        assert self._result is None
        self._result = result

    def safe_result(self) -> Result[Any, Exception]:
        assert self._result is not None
        return self._result

    def done(self) -> bool:
        return self._result is not None

    def next_child_name(self) -> str:
        return f"{self.id}.{self._num_children+1}"

    def create_child(
        self,
        id: str,
        invocation: LFI | RFI,
    ) -> Record[Any]:
        child_record = Record[Any](
            id=id,
            parent=self,
            invocation=invocation,
            ctx=self.ctx,
        )
        self.add_child(child_record)
        return child_record



from __future__ import annotations

import os
import time
from typing import TYPE_CHECKING, Any, Literal, final

import requests

from resonate.encoders import Base64Encoder, IEncoder
from resonate.errors import ResonateError
from resonate.logging import logger
from resonate.stores.record import (
    CallbackRecord,
    DurablePromiseRecord,
    InvokeMsg,
    ResumeMsg,
    TaskRecord,
)
from resonate.stores.traits import IPromiseStore

if TYPE_CHECKING:
    from collections.abc import Callable

    from resonate.typing import Data, Headers, IdempotencyKey, Tags


@final
class RemotePromiseStore(IPromiseStore):
    def __init__(
        self,
        url: str,
        call: Callable[[requests.Request], requests.Response],
        encoder: IEncoder[str, str],
    ) -> None:
        self.url = url
        self._call = call
        self._encoder = encoder

    def create(  # noqa: PLR0913
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        res = self._call(
            requests.Request(
                method="post",
                url=f"{self.url}/promises",
                headers=request_headers,
                json={
                    "id": id,
                    "param": {"headers": headers, "data": self._encode_data(data)},
                    "timeout": timeout,
                    "tags": tags,
                },
            )
        )

        return DurablePromiseRecord.decode(data=res.json(), encoder=self._encoder)

    def create_with_task(  # noqa: PLR0913
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
        pid: str,
        ttl: int,
    ) -> tuple[DurablePromiseRecord, TaskRecord | None]:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        res = self._call(
            requests.Request(
                method="post",
                url=f"{self.url}/promises/task",
                headers=request_headers,
                json={
                    "promise": {
                        "id": id,
                        "param": {"headers": headers, "data": self._encode_data(data)},
                        "timeout": timeout,
                        "tags": tags,
                    },
                    "task": {
                        "processId": pid,
                        "ttl": ttl,
                    },
                },
            )
        )

        data_json = res.json()
        task_data = data_json["task"]
        durable_promise = DurablePromiseRecord.decode(
            data_json["promise"], encoder=self._encoder
        )
        if task_data is None:
            return durable_promise, None
        return durable_promise, TaskRecord.decode(task_data, encoder=self._encoder)

    def resolve(
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        res = self._call(
            requests.Request(
                method="patch",
                url=f"{self.url}/promises/{id}",
                headers=request_headers,
                json={
                    "state": "RESOLVED",
                    "value": {"headers": headers, "data": self._encode_data(data)},
                },
            )
        )

        return DurablePromiseRecord.decode(data=res.json(), encoder=self._encoder)

    def reject(
        self,
        *,
        id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        res = self._call(
            requests.Request(
                method="patch",
                url=f"{self.url}/promises/{id}",
                headers=request_headers,
                json={
                    "state": "REJECTED",
                    "value": {"headers": headers, "data": self._encode_data(data)},
                },
            )
        )

        return DurablePromiseRecord.decode(data=res.json(), encoder=self._encoder)

    def cancel(
        self,
        *,
        id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord:
        request_headers = self._initialize_headers(strict=strict, ikey=ikey)

        res = self._call(
            requests.Request(
                method="patch",
                url=f"{self.url}/promises/{id}",
                headers=request_headers,
                json={
                    "state": "REJECTED_CANCELED",
                    "value": {"headers": headers, "data": self._encode_data(data)},
                },
            )
        )

        return DurablePromiseRecord.decode(data=res.json(), encoder=self._encoder)

    def get(self, *, id: str) -> DurablePromiseRecord:
        res = self._call(
            requests.Request(
                method="get",
                url=f"{self.url}/promises/{id}",
            )
        )

        return DurablePromiseRecord.decode(data=res.json(), encoder=self._encoder)

    def _initialize_headers(
        self, *, strict: bool, ikey: IdempotencyKey
    ) -> dict[str, str]:
        headers: dict[str, str] = {"Strict": str(strict)}
        if ikey is not None:
            headers["Idempotency-Key"] = ikey

        return headers

    def _encode_data(self, data: Data) -> str | None:
        if data is None:
            return None
        return _encode(data, self._encoder)


@final
class RemoteCallbackStore:
    def __init__(
        self,
        url: str,
        call: Callable[[requests.Request], requests.Response],
        encoder: IEncoder[str, str],
    ) -> None:
        self.url = url
        self._call = call
        self._encoder = encoder

    def create(
        self,
        *,
        id: str,
        promise_id: str,
        root_promise_id: str,
        timeout: int,
        recv: str | dict[str, Any],
    ) -> tuple[DurablePromiseRecord, CallbackRecord | None]:
        res = self._call(
            requests.Request(
                method="post",
                url=f"{self.url}/callbacks",
                json={
                    "id": id,
                    "promiseId": promise_id,
                    "rootPromiseId": root_promise_id,
                    "timeout": timeout,
                    "recv": recv,
                },
            )
        )

        data = res.json()
        callback_data = data["callback"]
        durable_promise = DurablePromiseRecord.decode(
            data["promise"], encoder=self._encoder
        )
        if callback_data is None:
            return durable_promise, None
        return durable_promise, CallbackRecord.decode(
            callback_data, encoder=self._encoder
        )


@final
class RemoteTaskStore:
    def __init__(
        self,
        url: str,
        call: Callable[[requests.Request], requests.Response],
        encoder: IEncoder[str, str],
    ) -> None:
        self.url = url
        self._call = call
        self._encoder = encoder

    def claim(
        self, *, task_id: str, counter: int, pid: str, ttl: int
    ) -> InvokeMsg | ResumeMsg:
        res = self._call(
            requests.Request(
                method="post",
                url=f"{self.url}/tasks/claim",
                json={
                    "id": task_id,
                    "counter": counter,
                    "processId": pid,
                    "ttl": ttl,
                },
            )
        )

        data = res.json()
        kind: Literal["resume", "invoke"] = data["type"]
        assert kind in ("resume", "invoke")

        promises: dict[str, Any] = data["promises"]
        if promises.get("leaf") is None:
            return InvokeMsg.decode(
                data=promises["root"]["data"], encoder=self._encoder
            )

        return ResumeMsg.decode(data=promises, encoder=self._encoder)

    def complete(self, *, task_id: str, counter: int) -> None:
        self._call(
            requests.Request(
                method="post",
                url=f"{self.url}/tasks/complete",
                json={
                    "id": task_id,
                    "counter": counter,
                },
            )
        )

    def heartbeat(self, *, pid: str) -> int:
        res = self._call(
            requests.Request(
                method="post",
                url=f"{self.url}/tasks/heartbeat",
                json={
                    "processId": pid,
                },
            )
        )

        return res.json()["tasksAffected"]


@final
class RemoteStore:
    def __init__(
        self,
        url: str | None = None,
        connect_timeout: int = 5,
        request_timeout: int = 5,
    ) -> None:
        self.url = (
            url
            if url is not None
            else os.getenv("RESONATE_SERVER", "http://localhost:8001")
        )
        self._encoder = Base64Encoder()
        self._session = requests.Session()
        self._timeout = (connect_timeout, request_timeout)

        self.promises = RemotePromiseStore(self.url, self._call, self._encoder)
        self.callbacks = RemoteCallbackStore(self.url, self._call, self._encoder)
        self.tasks = RemoteTaskStore(self.url, self._call, self._encoder)

    def _call(self, req: requests.Request) -> requests.Response:
        while True:
            try:
                res = self._session.send(req.prepare(), timeout=self._timeout)

                if res.ok:
                    return res

                # inspect the response and raise the corresponding error,
                # 5xx errors retry forever for now
                if res.status_code == requests.codes.bad_request:
                    msg = "Invalid request"
                    raise ResonateError(msg, "STORE_PAYLOAD", res.json())
                if res.status_code == requests.codes.unauthorized:
                    msg = "Unauthorized request"
                    raise ResonateError(msg, "STORE_UNAUTHORIZED", res.json())
                if res.status_code == requests.codes.forbidden:
                    msg = "Forbidden request"
                    raise ResonateError(msg, "STORE_FORBIDDEN", res.json())
                if res.status_code == requests.codes.not_found:
                    msg = "Not found"
                    raise ResonateError(msg, "STORE_NOT_FOUND", res.json())
                if res.status_code == requests.codes.conflict:
                    msg = "Already exists"
                    raise ResonateError(msg, "STORE_ALREADY_EXISTS", res.json())

                logger.warning(
                    "Unexpected response from remote store: %s",
                    res.json(),
                )
            except requests.RequestException as e:
                logger.warning("Unexpected error from remote store: %s", e)

            # wait for 1s and try again
            time.sleep(1)


def _encode(value: str, encoder: IEncoder[str, str]) -> str:
    try:
        return encoder.encode(value)
    except Exception as e:
        msg = "Encoder error"
        raise ResonateError(msg, "STORE_ENCODER", e) from e



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, TypedDict, TypeVar, overload
from uuid import uuid4

from typing_extensions import Concatenate, ParamSpec

from resonate.collections import FunctionRegistry
from resonate.dataclasses import RegisteredFn
from resonate.dependencies import Dependencies
from resonate.options import Options
from resonate.scheduler.scheduler import Scheduler
from resonate.scheduler.traits import IScheduler
from resonate.stores.remote import RemoteStore
from resonate.task_sources.poller import Poller
from resonate.task_sources.traits import ITaskSource

if TYPE_CHECKING:
    from collections.abc import Coroutine, Generator

    from resonate import retry_policy
    from resonate.context import Context
    from resonate.handle import Handle
    from resonate.scheduler.traits import IScheduler
    from resonate.stores.local import LocalStore
    from resonate.task_sources.traits import ITaskSource
    from resonate.typing import DurableCoro, DurableFn, Yieldable

P = ParamSpec("P")
T = TypeVar("T")


class _RunOptions(TypedDict):
    version: int


class Resonate:
    def __init__(
        self,
        pid: str | None = None,
        store: LocalStore | RemoteStore | None = None,
        task_source: ITaskSource | None = None,
    ) -> None:
        self._deps = Dependencies()
        self._registry = FunctionRegistry()

        self._scheduler: IScheduler = Scheduler(
            deps=self._deps,
            pid=pid or uuid4().hex,
            registry=self._registry,
            store=store or RemoteStore(),
            task_source=task_source or Poller(),
        )

        # start the scheduler
        self._scheduler.start()

    def stop(self) -> None:
        self._scheduler.stop()

    def set_dependency(self, key: str, obj: Any) -> None:  # noqa: ANN401
        self._deps.set(key, obj)

    @overload
    def register(
        self,
        func: Callable[
            Concatenate[Context, P],
            Generator[Yieldable, Any, T],
        ],
        name: str | None = None,
        version: int = 1,
        retry_policy: retry_policy.RetryPolicy | None = None,
    ) -> RegisteredFn[P, T]: ...
    @overload
    def register(
        self,
        func: Callable[Concatenate[Context, P], Coroutine[Any, Any, T]],
        name: str | None = None,
        version: int = 1,
        retry_policy: retry_policy.RetryPolicy | None = None,
    ) -> RegisteredFn[P, T]: ...
    @overload
    def register(
        self,
        func: Callable[Concatenate[Context, P], T],
        name: str | None = None,
        version: int = 1,
        retry_policy: retry_policy.RetryPolicy | None = None,
    ) -> RegisteredFn[P, T]: ...
    def register(
        self,
        func: DurableCoro[P, T] | DurableFn[P, T],
        name: str | None = None,
        version: int = 1,
        retry_policy: retry_policy.RetryPolicy | None = None,
    ) -> RegisteredFn[P, T]:
        if name is None:
            name = func.__name__
        self._registry.add(
            name,
            (func, Options(version=version, durable=True, retry_policy=retry_policy)),
        )
        return RegisteredFn[P, T](self._scheduler, func)

    @overload
    def run(
        self,
        id: str,
        func: Callable[
            Concatenate[Context, P],
            Generator[Yieldable, Any, T],
        ]
        | Callable[Concatenate[Context, P], T]
        | Callable[Concatenate[Context, P], Coroutine[Any, Any, T]],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Handle[T]: ...
    @overload
    def run(
        self,
        id: str,
        func: tuple[
            Callable[
                Concatenate[Context, P],
                Generator[Yieldable, Any, T],
            ]
            | Callable[Concatenate[Context, P], T]
            | Callable[Concatenate[Context, P], Coroutine[Any, Any, T]],
            _RunOptions,
        ],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Handle[T]: ...
    def run(
        self,
        id: str,
        func: Callable[
            Concatenate[Context, P],
            Generator[Yieldable, Any, T],
        ]
        | Callable[Concatenate[Context, P], T]
        | Callable[Concatenate[Context, P], Coroutine[Any, Any, T]]
        | tuple[
            Callable[
                Concatenate[Context, P],
                Generator[Yieldable, Any, T],
            ]
            | Callable[Concatenate[Context, P], T]
            | Callable[Concatenate[Context, P], Coroutine[Any, Any, T]],
            _RunOptions,
        ],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Handle[T]:
        if isinstance(func, tuple):
            raise NotImplementedError
        return self._scheduler.run(id, func, *args, **kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, TypeVar

from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from resonate.handle import Handle
    from resonate.typing import DurableCoro, DurableFn

P = ParamSpec("P")
T = TypeVar("T")


class IScheduler(ABC):
    @abstractmethod
    def start(self) -> None: ...

    @abstractmethod
    def stop(self) -> None: ...
    @abstractmethod
    def run(
        self,
        id: str,
        func: DurableCoro[P, T] | DurableFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Handle[T]: ...

    @abstractmethod
    def get(self, id: str) -> Handle[Any]: ...



from __future__ import annotations

import hashlib
import os
import traceback
import urllib
import urllib.parse
from functools import wraps
from importlib.metadata import version
from typing import Any, Callable, TypeVar
from uuid import UUID

from resonate.logging import logger

F = TypeVar("F", bound=Callable[..., Any])

T = TypeVar("T")


def string_to_uuid(string: str) -> str:
    return UUID(bytes=hashlib.sha1(string.encode("utf-8")).digest()[:16]).hex[-4:]  # noqa: S324


def exit_on_exception(func: F) -> F:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        try:
            return func(*args, **kwargs)
        except Exception:
            v = version("resonate-sdk")
            logger.exception(
                "An unexpected error happened.\n\nPlease report this issue so we can fix it as fast a possible:\n - https://github.com/resonatehq/resonate-sdk-py/issues/new?body=%s\n\n",  # noqa: E501
                urllib.parse.quote(
                    f"Resonate (version {v}) process exited with error:\n```bash\n{traceback.format_exc()}```"  # noqa: E501
                ),
            )
            os._exit(1)

    return wrapper  # type: ignore[return-value]



from __future__ import annotations

from typing import Literal

from typing_extensions import TypeAlias

ResonateErrorCode: TypeAlias = Literal[
    "UNKNOWN",
    "STORE_PAYLOAD",
    "STORE_UNAUTHORIZED",
    "STORE_FORBIDDEN",
    "STORE_NOT_FOUND",
    "STORE_ALREADY_EXISTS",
    "STORE_ENCODER",
]


class ResonateError(Exception):
    def __init__(
        self,
        msg: str,
        code: ResonateErrorCode,
        cause: Exception | None = None,
        *,
        retriable: bool = False,
    ) -> None:
        super().__init__(msg)
        self.msg = msg
        self.code = code
        self.cause = cause
        self.retriable = retriable

    @classmethod
    def from_error(cls, e: Exception) -> ResonateError:
        return ResonateError("Unknown error", "UNKNOWN", e, retriable=True)



from __future__ import annotations

import contextlib
import heapq
import time
from queue import Empty, Queue
from threading import Thread
from typing import Optional, TypeVar, final

from resonate import utils
from resonate.cmd_queue import CommandQ, Invoke
from resonate.traits import Subsystem

T = TypeVar("T")


def _ns_to_secs(ns: float) -> float:
    return ns / 1e9


def _secs_to_ns(secs: float) -> float:
    return secs * 1e9


@final
class DelayQueue(Subsystem):
    def __init__(self) -> None:
        self._inq = Queue[Optional[tuple[Invoke, float]]]()
        self._delayed: list[tuple[float, Invoke]] = []
        self._worker_thread: Thread | None = None

    def start(self, cmd_queue: CommandQ, pid: str) -> None:
        assert self._worker_thread is None, "Already been started."
        self._worker_thread = Thread(target=self._run, args=(cmd_queue,), daemon=True)
        self._worker_thread.start()

    def stop(self) -> None:
        assert self._worker_thread is not None, "Never started."
        self._inq.put(None)
        self._worker_thread.join()

    def _next_release_time(self) -> float:
        return self._delayed[0][0]

    @utils.exit_on_exception
    def _run(self, cmd_queue: CommandQ) -> None:
        """Worker thread that processes the delayed queue."""
        wait_time: float | None = None
        while True:
            current_time = time.time_ns()

            with contextlib.suppress(Empty):
                item_to_delay = self._inq.get(timeout=wait_time)
                if item_to_delay is None:
                    break
                item, delay = item_to_delay
                heapq.heappush(self._delayed, (current_time + _secs_to_ns(delay), item))
                self._inq.task_done()

            # Release any items whose delay has expired
            while self._delayed and self._next_release_time() <= current_time:
                _, item = heapq.heappop(self._delayed)
                cmd_queue.put(item)

            # Calculate the time to wait until the next item is ready
            if self._delayed:
                next_item_time = self._next_release_time()
                wait_time = _ns_to_secs(max(0, next_item_time - current_time))
            else:
                wait_time = None

    def enqueue(self, item: Invoke, delay: float) -> None:
        self._inq.put((item, delay))



from __future__ import annotations

from typing import TYPE_CHECKING, final

from resonate.logging import logger
from resonate.tracing import IAdapter

if TYPE_CHECKING:
    from resonate.events import SchedulerEvents


@final
class StdOutAdapter(IAdapter):
    def process_event(self, event: SchedulerEvents) -> None:
        logger.info(event)



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar, final, overload

from typing_extensions import ParamSpec, assert_never

from resonate.promise import Promise
from resonate.result import Err, Ok, Result

if TYPE_CHECKING:
    from collections.abc import Generator

    from resonate.handle import Handle
    from resonate.record import Record
    from resonate.scheduler.traits import IScheduler
    from resonate.typing import Data, DurableCoro, DurableFn, Headers, Tags, Yieldable

T = TypeVar("T")
P = ParamSpec("P")


@final
@dataclass(frozen=True)
class SQE(Generic[T]):
    thunk: Callable[[], T]
    id: str


@final
class RegisteredFn(Generic[P, T]):
    def __init__(
        self, scheduler: IScheduler, func: DurableCoro[P, T] | DurableFn[P, T], /
    ) -> None:
        self.fn = func
        self._scheduler = scheduler

    def run(self, id: str, *args: P.args, **kwargs: P.kwargs) -> Handle[T]:
        return self._scheduler.run(id, self.fn, *args, **kwargs)

    def get(self, id: str) -> Handle[T]:
        return self._scheduler.get(id)


@final
@dataclass(frozen=True)
class FinalValue(Generic[T]):
    v: Result[T, Exception]


class ResonateCoro(Generic[T]):
    def __init__(self, record: Record[T], coro: Generator[Yieldable, Any, T]) -> None:
        self.id = record.id
        self._coro = coro
        self._coro_active: bool = True
        self._final_value: Result[T, Exception] | None = None
        self._first_error_used = False
        self._next_child_to_yield = 0
        self._record = record

    def _pending_children(self) -> list[Record[Any]]:
        return self._record.children[self._next_child_to_yield :]

    def advance(
        self, next_value: Result[Any, Exception] | None
    ) -> Yieldable | FinalValue[T]:
        yielded: Yieldable
        try:
            if next_value is None:
                # `None` means to initialize the coroutine
                assert self._coro_active
                yielded = next(self._coro)
            elif isinstance(next_value, Ok):
                # `Ok[T]` means send a value
                yielded = self._send(next_value.unwrap())
            elif isinstance(next_value, Err):
                # `Err[Exception]` means throw and Exception
                yielded = self._throw(next_value.err())
            else:
                assert_never(next_value)
        except StopIteration as e:
            return FinalValue(Ok(e.value))
        except Exception as e:  # noqa: BLE001
            return FinalValue(Err(e))
        return yielded

    def _send(self, v: T) -> Yieldable:
        if self._coro_active:
            try:
                return self._coro.send(v)
            except StopIteration as e:
                assert self._final_value is None, "return value can only be set once."
                self._final_value = Ok(e.value)
                self._coro_active = False

        for child in self._pending_children():
            self._next_child_to_yield += 1
            if child.done():
                continue
            return Promise[Any](child.id)

        assert all(child.done() for child in self._record.children)
        assert not self._coro_active
        assert isinstance(self._final_value, Ok)
        raise StopIteration(self._final_value.unwrap())

    def _throw(self, error: Exception) -> Yieldable:
        if self._coro_active:
            try:
                return self._coro.throw(error)
            except StopIteration as e:
                assert self._final_value is None, "return value can only be set once."
                self._final_value = Ok(e.value)
                self._coro_active = False

        if not self._first_error_used:
            self._final_value = Err(error)
            self._first_error_used = True

        for child in self._pending_children():
            self._next_child_to_yield += 1
            if child.done():
                continue
            return Promise[Any](child.id)
        assert all(
            child.done() for child in self._record.children
        ), "All children promise must have been resolved."

        assert not self._coro_active
        assert self._final_value is not None
        if isinstance(self._final_value, Ok):
            raise StopIteration(self._final_value)
        if isinstance(self._final_value, Err):
            raise self._final_value.err()
        assert_never(self._final_value)


class Invocation(Generic[T]):
    @overload
    def __init__(
        self,
        fn: str,
        /,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> None: ...
    @overload
    def __init__(
        self,
        fn: DurableCoro[P, T] | DurableFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None: ...
    def __init__(
        self,
        fn: DurableCoro[P, T] | DurableFn[P, T] | str,
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.fn = fn
        self.args = args
        self.kwargs = kwargs


class DurablePromise:
    def __init__(
        self,
        id: str | None = None,
        data: Data = None,
        headers: Headers = None,
        tags: Tags = None,
        timeout: int = sys.maxsize,
    ) -> None:
        self.id = id
        self.data = data
        self.headers = headers
        self.tags = tags
        self.timeout = timeout



from __future__ import annotations

import json
import os
import random
import time
from functools import cache, lru_cache
from typing import TYPE_CHECKING, Any

import pytest

from resonate import DurablePromise, Handle, Resonate
from resonate.promise import Promise
from resonate.retry_policy import constant, exponential, linear, never
from resonate.stores import LocalStore, RemoteStore
from resonate.targets import poll
from resonate.task_sources import Poller

if TYPE_CHECKING:
    from collections.abc import Generator

    from resonate.context import Context
    from resonate.typing import Yieldable


@lru_cache
def _fib(n: int) -> int:
    if n <= 1:
        return n
    return _fib(n - 1) + _fib(n - 2)


@lru_cache
def _factorial(n: int) -> int:
    if n == 0:
        return 1
    return n * _factorial(n - 1)


@cache
def _promise_storages() -> list[LocalStore | RemoteStore]:
    stores: list[LocalStore | RemoteStore] = [LocalStore()]
    if os.getenv("RESONATE_STORE_URL") is not None:
        stores.append(RemoteStore(url=os.environ["RESONATE_STORE_URL"]))
    return stores


@pytest.mark.parametrize("store", _promise_storages())
def test_calling_only_sync_function(store: LocalStore | RemoteStore) -> None:
    def foo_sync(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(store=store)
    resonate.register(foo_sync)
    p: Handle[str] = resonate.run("foo-sync", foo_sync, "hi")
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_calling_only_async_function(store: LocalStore | RemoteStore) -> None:
    async def foo_async(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(store=store)
    resonate.register(foo_async)
    p: Handle[str] = resonate.run("foo-async", foo_async, "hi")
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_golden_device_lfi(store: LocalStore | RemoteStore) -> None:
    def foo_golden_device_lfi(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.lfi(bar_golden_device_lfi, n).options(
            durable=random.choice([True, False])  # noqa: S311
        )
        assert isinstance(p, Promise)
        v: str = yield p
        assert isinstance(v, str)
        return v

    def bar_golden_device_lfi(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(store=store)
    resonate.register(foo_golden_device_lfi)
    p: Handle[str] = resonate.run("test-golden-device-lfi", foo_golden_device_lfi, "hi")
    assert isinstance(p, Handle)
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_factorial_lfi(store: LocalStore | RemoteStore) -> None:
    def exec_id(n: int) -> str:
        return f"test-factorial-lfi-{n}"

    def factorial_lfi(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        p = yield ctx.lfi(factorial_lfi, n - 1).options(
            id=exec_id(n - 1),
            durable=random.choice([True, False]),  # noqa: S311
        )
        return n * (yield p)

    resonate = Resonate(store=store)
    resonate.register(factorial_lfi)
    n = 30
    p: Handle[int] = resonate.run(exec_id(n), factorial_lfi, n)
    assert p.result() == _factorial(n)
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_fibonacci_preorder_lfi(store: LocalStore | RemoteStore) -> None:
    def exec_id(n: int) -> str:
        return f"test-fib-preorder-lfi-{n}"

    def fib_lfi(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n
        p1 = yield ctx.lfi(fib_lfi, n - 1).options(
            id=exec_id(n - 1),
            durable=random.choice([True, False]),  # noqa: S311
        )
        p2 = yield ctx.lfi(fib_lfi, n - 2).options(
            id=exec_id(n - 2),
            durable=random.choice([True, False]),  # noqa: S311
        )
        n1 = yield p1
        n2 = yield p2
        return n1 + n2

    resonate = Resonate(store=store)
    resonate.register(fib_lfi)
    n = 30
    p: Handle[int] = resonate.run(exec_id(n), fib_lfi, n)
    assert p.result() == _fib(n)
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_fibonacci_postorder_lfi(store: LocalStore | RemoteStore) -> None:
    def exec_id(n: int) -> str:
        return f"test-fib-postorder-lfi-{n}"

    def fib_lfi(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n
        p1 = yield ctx.lfi(fib_lfi, n - 1).options(
            id=exec_id(n - 1),
            durable=random.choice([True, False]),  # noqa: S311
        )
        p2 = yield ctx.lfi(fib_lfi, n - 2).options(
            id=exec_id(n - 2),
            durable=random.choice([True, False]),  # noqa: S311
        )
        n2 = yield p2
        n1 = yield p1
        return n1 + n2

    resonate = Resonate(store=store)
    resonate.register(fib_lfi)
    n = 30
    p: Handle[int] = resonate.run(exec_id(n), fib_lfi, n)
    assert p.result() == _fib(n)
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_golden_device_lfc(store: RemoteStore | LocalStore) -> None:
    def foo_golden_device_lfc(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.lfc(bar_golden_device_lfi, n).options(
            durable=random.choice([True, False])  # noqa: S311
        )
        assert isinstance(v, str)
        return v

    def bar_golden_device_lfi(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(store=store)
    resonate.register(foo_golden_device_lfc)
    p: Handle[str] = resonate.run("test-golden-device-lfc", foo_golden_device_lfc, "hi")
    assert isinstance(p, Handle)
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_factorial_lfc(store: LocalStore | RemoteStore) -> None:
    def exec_id(n: int) -> str:
        return f"test-factorial-lfc-{n}"

    def factorial_lfc(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1

        return n * (
            yield ctx.lfc(factorial_lfc, n - 1).options(
                id=exec_id(n - 1),
                durable=random.choice([True, False]),  # noqa: S311
            )
        )

    resonate = Resonate(store=store)
    resonate.register(factorial_lfc)
    n = 30
    p: Handle[int] = resonate.run(exec_id(n), factorial_lfc, n)
    assert p.result() == _factorial(n)
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_fibonacci_lfc(store: LocalStore | RemoteStore) -> None:
    def exec_id(n: int) -> str:
        return f"test-fib-lfc-{n}"

    def fib_lfc(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n
        n1 = yield ctx.lfc(fib_lfc, n - 1).options(
            id=exec_id(n - 1),
            durable=random.choice([True, False]),  # noqa: S311
        )
        n2 = yield ctx.lfc(fib_lfc, n - 2).options(
            id=exec_id(n - 2),
            durable=random.choice([True, False]),  # noqa: S311
        )
        return n1 + n2

    resonate = Resonate(store=store)
    resonate.register(fib_lfc)
    n = 30
    p: Handle[int] = resonate.run(exec_id(n), fib_lfc, n)
    assert p.result() == _fib(n)
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_golden_device_rfi() -> None:
    group = "test-golden-device-rfi"

    def foo_golden_device_rfi(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.rfi(bar_golden_device_rfi, n).options(
            id="bar", send_to=poll(group)
        )
        assert isinstance(p, Promise)
        v: str = yield p
        assert isinstance(v, str)
        return v

    def bar_golden_device_rfi(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(foo_golden_device_rfi)
    resonate.register(bar_golden_device_rfi)
    p: Handle[str] = resonate.run(f"{group}-foo", foo_golden_device_rfi, "hi")
    assert isinstance(p, Handle)
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_factorial_rfi() -> None:
    group = "test-factorial-rfi"

    def exec_id(n: int) -> str:
        return f"test-factorial-rfi-{n}"

    def factorial_rfi(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        p = yield ctx.rfi(factorial_rfi, n - 1).options(
            id=exec_id(n - 1), send_to=poll(group)
        )
        return n * (yield p)

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(factorial_rfi)
    n = 10
    p: Handle[int] = resonate.run(exec_id(n), factorial_rfi, n)
    assert p.result() == _factorial(n)
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_fibonacci_preorder_rfi() -> None:
    group = "test-fibonacci-preorder-rfi"

    def exec_id(n: int) -> str:
        return f"test-fib-preorder-rfi-{n}"

    def fib_rfi(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n
        p1 = yield ctx.rfi(fib_rfi, n - 1).options(
            id=exec_id(n - 1), send_to=poll(group)
        )
        n1 = yield p1
        p2 = yield ctx.rfi(fib_rfi, n - 2).options(
            id=exec_id(n - 2), send_to=poll(group)
        )
        n2 = yield p2
        return n1 + n2

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(fib_rfi)
    n = 10
    p: Handle[int] = resonate.run(exec_id(n), fib_rfi, n)
    assert p.result() == _fib(n)
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_fibonacci_postorder_rfi() -> None:
    group = "test-fibonacci-postorder-rfi"

    def exec_id(n: int) -> str:
        return f"test-fib-postorder-rfi-{n}"

    def fib_rfi(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n
        p1 = yield ctx.rfi(fib_rfi, n - 1).options(
            id=exec_id(n - 1), send_to=poll(group)
        )
        p2 = yield ctx.rfi(fib_rfi, n - 2).options(
            id=exec_id(n - 2), send_to=poll(group)
        )
        n1 = yield p1
        n2 = yield p2
        return n1 + n2

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(fib_rfi)
    n = 10

    p: Handle[int] = resonate.run(exec_id(n), fib_rfi, n)
    assert p.result() == _fib(n)
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_golden_device_rfi_and_lfc() -> None:
    group = "test-golden-device-rfi-and-lfc"

    def foo(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.lfc(bar, n).options(
            id="bar",
            durable=False,
        )
        return v

    def bar(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.rfi(baz, n).options(id="baz", send_to=poll(group))
        v: str = yield p
        return v

    def baz(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(foo)
    resonate.register(baz)
    p: Handle[str] = resonate.run(f"{group}-foo", foo, "hi")
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.skip
@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_fibonacci_full_random() -> None:
    rand = random.Random(444)  # noqa: S311

    def exec_id(n: int) -> str:
        return f"fib({n})"

    def fib(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n

        if rand.choice((True, False)):
            p1 = yield ctx.rfi(fib, n - 1).options(
                id=exec_id(n - 1),
            )
        else:
            p1 = yield ctx.lfi(fib, n - 1).options(
                id=exec_id(n - 1),
                durable=rand.choice((True, False)),
            )

        if rand.choice((True, False)):
            p2 = yield ctx.rfi(fib, n - 2).options(
                id=exec_id(n - 2),
            )
        else:
            p2 = yield ctx.lfi(fib, n - 2).options(
                id=exec_id(n - 2),
                durable=rand.choice((True, False)),
            )

        promises = [p1, p2]
        total: int = 0
        while promises:
            idx = rand.randint(0, len(promises) - 1)
            p = promises.pop(idx)
            v = yield p
            total += v

        return total

    resonate = Resonate(store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]))
    resonate.register(fib)
    n = 3
    p: Handle[int] = resonate.run(exec_id(n), fib, n)
    assert p.result() == _fib(n)
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_golden_device_rfi_and_lfc_with_decorator() -> None:
    group = "test-golden-device-rfi-and-lfc-with-decorator"

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )

    @resonate.register
    def foo(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.lfc(bar, n).options(
            id="bar",
            durable=False,
        )
        return v

    def bar(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        p: Promise[str] = yield ctx.rfi(baz, n).options(id="baz", send_to=poll(group))
        v: str = yield p
        return v

    @resonate.register
    def baz(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    p: Handle[str] = foo.run(f"{group}-foo", n="hi")
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_golden_device_rfc() -> None:
    group = "test-golden-device-rfc"

    def foo_golden_device_rfc(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.rfc(bar_golden_device_rfc, n).options(
            id="bar", send_to=poll(group)
        )
        assert isinstance(v, str)
        return v

    def bar_golden_device_rfc(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(foo_golden_device_rfc)
    resonate.register(bar_golden_device_rfc)
    p: Handle[str] = resonate.run(f"{group}-foo", foo_golden_device_rfc, "hi")
    assert isinstance(p, Handle)
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_factorial_rfc() -> None:
    group = "test-factorial-rfc"

    def exec_id(n: int) -> str:
        return f"test-factorial-rfc-{n}"

    def factorial_rfc(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n == 0:
            return 1
        return n * (
            yield ctx.rfc(factorial_rfc, n - 1).options(
                id=exec_id(n - 1), send_to=poll(group)
            )
        )

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(factorial_rfc)
    n = 10
    p: Handle[int] = resonate.run(exec_id(n), factorial_rfc, n)
    assert p.result() == _factorial(n)
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_fibonacci_preorder_rfc() -> None:
    group = "test-fibonacci-preorder-rfc"

    def exec_id(n: int) -> str:
        return f"test-fib-preorder-rfc-{n}"

    def fib_rfc(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        if n <= 1:
            return n
        n1 = yield ctx.rfc(fib_rfc, n - 1).options(
            id=exec_id(n - 1), send_to=poll(group)
        )
        n2 = yield ctx.rfc(fib_rfc, n - 2).options(
            id=exec_id(n - 2), send_to=poll(group)
        )
        return n1 + n2

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(fib_rfc)
    n = 10
    p: Handle[int] = resonate.run(exec_id(n), fib_rfc, n)
    assert p.result() == _fib(n)
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_golden_device_rfc_and_lfc() -> None:
    group = "test-golden-device-rfc-and-lfc"

    def foo(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.lfc(bar, n).options(
            id="bar",
            durable=False,
        )
        return v

    def bar(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.rfc(baz, n).options(id="baz", send_to=poll(group))
        return v

    def baz(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )
    resonate.register(foo)
    resonate.register(baz)
    p: Handle[str] = resonate.run(f"{group}-foo", foo, "hi")
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_golden_device_rfc_and_lfc_with_decorator() -> None:
    group = "test-golden-device-rfc-and-lfc-with-decorator"

    resonate = Resonate(
        store=RemoteStore(url=os.environ["RESONATE_STORE_URL"]),
        task_source=Poller("http://localhost:8002", group=group),
    )

    @resonate.register
    def foo(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.lfc(bar, n).options(
            id="bar",
            durable=False,
        )
        return v

    def bar(ctx: Context, n: str) -> Generator[Yieldable, Any, str]:
        v: str = yield ctx.rfc(baz, n).options(id="baz", send_to=poll(group))
        return v

    @resonate.register
    def baz(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    p: Handle[str] = foo.run(f"{group}-foo", n="hi")
    assert p.result() == "hi"
    resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_retry_policies_local_coroutine(store: LocalStore | RemoteStore) -> None:
    def foo_retry_policy(ctx: Context, n: str) -> Generator[Yieldable, str, str]:
        yield ctx.lfc(bar_retry_policy, n)
        msg = "oops something went wrong"
        raise RuntimeError(msg)

    def bar_retry_policy(ctx: Context, n: str) -> str:  # noqa: ARG001
        return n

    for policy in (
        never(),
        constant(delay=0, max_retries=3),
        linear(delay=0, max_retries=3),
        exponential(base_delay=0, factor=1, max_retries=1, max_delay=5),
    ):
        resonate = Resonate(store=store)
        resonate.register(foo_retry_policy, retry_policy=policy)
        p: Handle[str] = resonate.run(
            f"foo-retry-policy-local-coroutine-{policy!s}", foo_retry_policy, "hi"
        )
        with pytest.raises(RuntimeError):
            p.result()
        resonate.stop()


@pytest.mark.parametrize("store", _promise_storages())
def test_retry_policies_local_func(store: LocalStore | RemoteStore) -> None:
    def foo_retry_policy(ctx: Context, n: str) -> str:  # noqa: ARG001
        msg = "oops something went wrong"
        raise RuntimeError(msg)

    for policy in (
        never(),
        constant(delay=0, max_retries=3),
        linear(delay=0, max_retries=3),
        exponential(base_delay=0, factor=1, max_retries=1, max_delay=5),
    ):
        resonate = Resonate(store=store)
        resonate.register(foo_retry_policy, retry_policy=policy)
        p: Handle[str] = resonate.run(
            f"foo-retry-policy-local-func-{policy!s}", foo_retry_policy, "hi"
        )
        with pytest.raises(RuntimeError):
            p.result()
        resonate.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_human_in_the_loop() -> None:
    group = "test-human-in-the-loop"

    def _user_manual_completion(id: str) -> DurablePromise:
        return DurablePromise(id=id)

    def human_in_the_loop(ctx: Context) -> Generator[Yieldable, Any, str]:
        name: str = yield ctx.rfc(
            _user_manual_completion("test-human-in-loop-question-to-answer-1")
        )
        age: int = yield ctx.rfc(
            _user_manual_completion(id="test-human-in-loop-question-to-answer-2")
        )
        return f"Hi {name} with age {age}"

    store = RemoteStore(url=os.environ["RESONATE_STORE_URL"])
    s = Resonate(store=store, task_source=Poller("http://localhost:8002", group=group))
    s.register(human_in_the_loop)
    p: Handle[str] = s.run("test-feature-human-in-the-loop", human_in_the_loop)
    time.sleep(3)
    store.promises.resolve(
        id="test-human-in-loop-question-to-answer-1",
        ikey=None,
        strict=False,
        headers=None,
        data=json.dumps("Peter"),
    )
    time.sleep(3)
    store.promises.resolve(
        id="test-human-in-loop-question-to-answer-2",
        ikey=None,
        strict=False,
        headers=None,
        data=json.dumps(50),
    )
    time.sleep(3)
    assert p.result() == "Hi Peter with age 50"
    s.stop()


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_sleep() -> None:
    group = "test-sleep"

    store = RemoteStore(url="http://localhost:8001")
    s = Resonate(store=store, task_source=Poller("http://localhost:8002", group=group))

    @s.register
    def foo_sleep(ctx: Context, n: int) -> Generator[Yieldable, Any, int]:
        yield ctx.sleep(n)
        return n

    n = 1
    p = foo_sleep.run(f"{group}-{n}", n)
    assert p.result() == n
    s.stop()



from __future__ import annotations

from typing import Any

from resonate.actions import LFI, RFI
from resonate.context import Context
from resonate.dataclasses import Invocation
from resonate.dependencies import Dependencies
from resonate.record import Record


def test_lineage() -> None:
    lfi = LFI(Invocation("foo", 1, 2))
    rfi = RFI(Invocation("foo", 1, 2))
    n0 = Record[Any](id="n0", invocation=lfi, parent=None, ctx=Context(Dependencies()))
    assert n0.is_root
    assert n0.root() == n0
    assert n0.children == []
    n1 = n0.create_child("n1", rfi)
    assert n0.children == [n1]
    n2 = n0.create_child("n2", rfi)
    assert n0.children == [n1, n2]
    assert n1.root() == n1
    assert n2.root() == n2
    assert n1.children == []
    assert n2.children == []

    n3 = n1.create_child("n3", lfi)
    n4 = n1.create_child("n4", lfi)
    n5 = n2.create_child("n5", lfi)
    n6 = n2.create_child("n6", lfi)
    assert not n3.is_root
    assert not n4.is_root
    assert not n5.is_root
    assert not n6.is_root
    assert n3.root() == n1
    assert n4.root() == n1
    assert n5.root() == n2
    assert n6.root() == n2

    assert n1.children == [n3, n4]
    assert n2.children == [n5, n6]
    assert n0.children == [n1, n2]


def test_lineage_2() -> None:
    lfi = LFI(Invocation("foo", 1, 2))
    rfi = RFI(Invocation("foo", 1, 2))
    foo = Record[Any](
        id="foo", invocation=rfi, parent=None, ctx=Context(Dependencies())
    )
    bar = foo.create_child(id="bar", invocation=lfi)
    bar.create_child(id="baz", invocation=rfi)
    assert foo.children == [bar]



from __future__ import annotations

import json
import os
import sys
import uuid

import pytest

from resonate.stores.remote import (
    RemoteStore,
)


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_0_callback_on_existing_promise() -> None:
    store = RemoteStore(url=os.environ["RESONATE_STORE_URL"])
    id = "0.0"
    store.promises.create(
        id=id,
        ikey=None,
        strict=False,
        headers=None,
        data=None,
        timeout=sys.maxsize,
        tags=None,
    )
    promise_record, callback_record = store.callbacks.create(
        id=id + id,
        promise_id=id,
        root_promise_id=id,
        timeout=sys.maxsize,
        recv="default",
    )
    assert callback_record is not None
    assert promise_record.id == callback_record.id
    assert callback_record.timeout == sys.maxsize
    assert callback_record.id == id


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_1_callback_on_non_existing_promise() -> None:
    store = RemoteStore(url=os.environ["RESONATE_STORE_URL"])
    id = "1.1"
    with pytest.raises(Exception):  # noqa: B017, PT011
        store.callbacks.create(
            id=id + id,
            promise_id=id,
            root_promise_id=id,
            timeout=sys.maxsize,
            recv="default",
        )


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_2_callback_on_resolved_promise() -> None:
    store = RemoteStore(url=os.environ["RESONATE_STORE_URL"])
    id = "2.0"
    store.promises.create(
        id=id,
        ikey=None,
        strict=False,
        headers=None,
        data=None,
        timeout=sys.maxsize,
        tags=None,
    )
    store.promises.resolve(id=id, ikey=None, strict=False, headers=None, data="1")
    promise_record, callback_record = store.callbacks.create(
        id=id + id,
        promise_id=id,
        root_promise_id=id,
        timeout=sys.maxsize,
        recv="default",
    )
    assert promise_record.is_completed()
    assert callback_record is None


@pytest.mark.skipif(
    os.getenv("RESONATE_STORE_URL") is None, reason="env variable is not set"
)
def test_case_3_create_durable_promise_with_task() -> None:
    store = RemoteStore(url=os.environ["RESONATE_STORE_URL"])
    pid = uuid.uuid4().hex
    ttl = 5 * 1000
    id = "3.0"
    durable_promise, task_record = store.promises.create_with_task(
        id=id,
        ikey=None,
        strict=False,
        headers=None,
        data=None,
        timeout=sys.maxsize,
        tags={
            "resonate:invoke": json.dumps(
                {"type": "poll", "data": {"group": "default", "id": pid}}
            )
        },
        pid=pid,
        ttl=ttl,
    )
    assert durable_promise.id == id
    assert task_record is not None
    assert task_record.counter == 1
    store.tasks.heartbeat(pid=pid)
    store.tasks.complete(task_id=task_record.task_id, counter=task_record.counter)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from resonate.traits import Subsystem


class ITaskSource(Subsystem, ABC):
    @abstractmethod
    def default_recv(self, pid: str) -> dict[str, Any]: ...



from __future__ import annotations

import asyncio
import contextlib
import json
import sys
import time
from functools import partial
from inspect import iscoroutinefunction, isfunction, isgeneratorfunction
from queue import Queue
from threading import Thread
from typing import TYPE_CHECKING, Any, TypeVar

import requests
from typing_extensions import ParamSpec, assert_never

from resonate import utils
from resonate.actions import DI, LFC, LFI, RFC, RFI
from resonate.cmd_queue import (
    Claim,
    Command,
    CommandQ,
    Complete,
    ForkOrJoin,
    Invoke,
    Notify,
    Resume,
    Subscribe,
)
from resonate.context import Context
from resonate.dataclasses import (
    SQE,
    DurablePromise,
    FinalValue,
    Invocation,
    ResonateCoro,
)
from resonate.delay_queue import DelayQueue
from resonate.encoders import JsonEncoder
from resonate.handle import Handle
from resonate.logging import logger
from resonate.processor.processor import Processor
from resonate.promise import Promise
from resonate.record import Record
from resonate.result import Err, Ok, Result
from resonate.scheduler.traits import IScheduler
from resonate.stores.record import (
    DurablePromiseRecord,
    InvokeMsg,
    ResumeMsg,
    TaskRecord,
)
from resonate.stores.remote import RemoteStore

if TYPE_CHECKING:
    from resonate.collections import FunctionRegistry
    from resonate.dependencies import Dependencies
    from resonate.stores.local import LocalStore
    from resonate.stores.record import TaskRecord
    from resonate.task_sources.traits import ITaskSource
    from resonate.typing import Data, DurableCoro, DurableFn, Headers, Tags

T = TypeVar("T")
P = ParamSpec("P")


class Scheduler(IScheduler):
    def __init__(
        self,
        deps: Dependencies,
        pid: str,
        registry: FunctionRegistry,
        store: LocalStore | RemoteStore,
        task_source: ITaskSource,
    ) -> None:
        self._deps = deps
        self._pid = pid
        self._processor = Processor()
        self._registry = registry
        self._store = store
        self._task_source = task_source

        self._cmd_queue: CommandQ = Queue()

        self._awaiting_rfi: dict[str, list[str]] = {}
        self._awaiting_lfi: dict[str, list[str]] = {}
        self._subsriptions: dict[str, list[Handle[Any]]] = {}
        self._records: dict[str, Record[Any]] = {}

        self._encoder = JsonEncoder()
        self._recv = self._task_source.default_recv(self._pid)

        self._delay_queue = DelayQueue()

        self._heartbeat_thread = Thread(target=self._heartbeat, daemon=True)
        self._scheduler_thread = Thread(target=self._loop, daemon=True)

    def start(self) -> None:
        if isinstance(self._store, RemoteStore):
            # start the heartbeat thread
            self._heartbeat_thread.start()

            # start the task source
            self._task_source.start(self._cmd_queue, self._pid)

        # start delay queue
        self._delay_queue.start(self._cmd_queue, self._pid)

        # start the processor
        self._processor.start(self._cmd_queue, self._pid)

        # start the scheduler
        self._scheduler_thread.start()

    def get(self, id: str) -> Handle[Any]:
        handle = Handle[Any](id)
        self._cmd_queue.put(Subscribe(id, handle))
        return handle

    def run(
        self,
        id: str,
        func: DurableCoro[P, T] | DurableFn[P, T],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Handle[T]:
        # If there's already a record with this ID, dedup.
        handle = Handle[T](id)
        self._cmd_queue.put(ForkOrJoin(id, handle, Invocation(func, *args, **kwargs)))
        return handle

    @utils.exit_on_exception
    def _heartbeat(self) -> None:
        assert isinstance(self._store, RemoteStore)
        while True:
            affected: int | None = None
            with contextlib.suppress(requests.exceptions.ConnectionError):
                affected = self._store.tasks.heartbeat(pid=self._pid)
                logger.debug("Heatbeat affected %s tasks", affected)
            time.sleep(2)

    @utils.exit_on_exception
    def _loop(self) -> None:
        while True:
            cmd = self._cmd_queue.get()
            if cmd is None:
                break

            # start the next tick
            for loopback in self._step(cmd):
                self._cmd_queue.put(loopback)

            # mark task done
            self._cmd_queue.task_done()

    def _step(self, cmd: Command) -> list[Command]:  # noqa: PLR0911
        if isinstance(cmd, Invoke):
            return self._handle_invoke(cmd)
        if isinstance(cmd, Resume):
            return self._handle_resume(cmd)
        if isinstance(cmd, Complete):
            return self._handle_complete(cmd)
        if isinstance(cmd, Claim):
            return self._handle_claim(cmd)
        if isinstance(cmd, ForkOrJoin):
            return self._handle_fork_or_join(cmd)
        if isinstance(cmd, Subscribe):
            return self._handle_subscribe(cmd)
        if isinstance(cmd, Notify):
            return self._handle_notify(cmd)
        assert_never(cmd)

    def _handle_invoke(self, invoke: Invoke) -> list[Command]:
        logger.info("Ingesting record %s", invoke.id)
        record = self._records[invoke.id]
        assert not record.done()
        assert isinstance(record.invocation.unit, Invocation)
        assert not isinstance(record.invocation.unit.fn, str)
        fn, args, kwargs = (
            record.invocation.unit.fn,
            record.invocation.unit.args,
            record.invocation.unit.kwargs,
        )
        if isgeneratorfunction(fn):
            record.add_coro(
                ResonateCoro(
                    record,
                    fn(
                        record.ctx,
                        *args,
                        **kwargs,
                    ),
                )
            )
            return self._handle_continue(record.id, next_value=None)

        if iscoroutinefunction(fn):
            self._processor.enqueue(
                SQE[Any](
                    thunk=partial(
                        asyncio.run,
                        fn(
                            record.ctx,
                            *args,
                            **kwargs,
                        ),
                    ),
                    id=record.id,
                )
            )

        else:
            assert isfunction(fn)
            self._processor.enqueue(
                SQE[Any](
                    thunk=partial(
                        fn,
                        record.ctx,
                        *args,
                        **kwargs,
                    ),
                    id=record.id,
                )
            )

        return []

    def _handle_resume(self, resume: Resume) -> list[Command]:
        return self._handle_continue(resume.id, resume.next_value)

    def _handle_complete(
        self,
        complete: Complete,
    ) -> list[Command]:
        record = self._records[complete.id]
        value = complete.result
        if isinstance(record, str):
            record = self._records[record]
        if record.should_retry(value):
            record.increate_attempt()
            self._delay_queue.enqueue(Invoke(record.id), record.next_retry_delay())
            return []

        loopback: list[Command] = []
        if record.invocation.opts.durable:
            durable_promise: DurablePromiseRecord
            if isinstance(value, Ok):
                durable_promise = self._store.promises.resolve(
                    id=record.id,
                    ikey=utils.string_to_uuid(record.id),
                    strict=False,
                    headers=None,
                    data=self._encoder.encode(value.unwrap()),
                )

            elif isinstance(value, Err):
                durable_promise = self._store.promises.reject(
                    id=record.id,
                    ikey=utils.string_to_uuid(record.id),
                    strict=False,
                    headers=None,
                    data=self._encoder.encode(value.err()),
                )
            else:
                assert_never(value)

            value = durable_promise.get_value(self._encoder)
            assert not record.done()

            if record.has_task():
                self._complete_task(record.id)

            loopback.append(Notify(complete.id, value))
            record.set_result(value, deduping=False)
            loopback.extend(self._unblock_awaiting_local(record.id))

            root = record.root()
            if root != record and self._blocked_only_on_remote(root.id):
                self._complete_task(root.id)
            return loopback

        loopback.append(Notify(complete.id, value))
        record.set_result(value, deduping=False)
        loopback.extend(self._unblock_awaiting_local(record.id))
        return loopback

    def _handle_claim(self, claim: Claim) -> list[Command]:
        assert isinstance(self._store, RemoteStore)
        invoke_or_resume = self._store.tasks.claim(
            task_id=claim.record.task_id,
            counter=claim.record.counter,
            pid=self._pid,
            ttl=5 * 1000,
        )
        if isinstance(invoke_or_resume, InvokeMsg):
            return self._process_invoke_msg(invoke_or_resume, claim.record)
        if isinstance(invoke_or_resume, ResumeMsg):
            return self._process_resume_msg(invoke_or_resume, claim.record)

        assert_never(invoke_or_resume)

    def _handle_fork_or_join(self, fork_or_join: ForkOrJoin) -> list[Command]:
        record = self._records.get(fork_or_join.id)
        if record:
            if record.done():
                fork_or_join.handle.set_result(record.safe_result())
            else:
                return [Subscribe(fork_or_join.id, fork_or_join.handle)]

        # Get function name from registry
        assert not isinstance(fork_or_join.invocation.fn, str)
        fn_name = self._registry.get_from_value(fork_or_join.invocation.fn)
        assert (
            fn_name is not None
        ), f"Function {fork_or_join.invocation.fn.__name__} must be registered"
        func_with_options = self._registry.get(fn_name)
        assert func_with_options is not None
        opts = func_with_options[-1]
        assert opts.durable, "Top level must always be durable."

        record = Record[Any](
            id=fork_or_join.id,
            parent=None,
            invocation=RFI(fork_or_join.invocation, opts),
            ctx=Context(self._deps),
        )
        self._records[record.id] = record

        # Create durable promise while claiming the task.
        assert self._recv
        durable_promise, task = self._store.promises.create_with_task(
            id=fork_or_join.id,
            ikey=utils.string_to_uuid(fork_or_join.id),
            strict=False,
            headers=None,
            data=self._encoder.encode(
                {
                    "func": fn_name,
                    "args": fork_or_join.invocation.args,
                    "kwargs": fork_or_join.invocation.kwargs,
                }
            ),
            timeout=sys.maxsize,
            tags={"resonate:invoke": json.dumps(self._recv)},
            pid=self._pid,
            ttl=5 * 1_000,
        )

        record.add_durable_promise(durable_promise)
        if task:
            record.add_task(task)

        if durable_promise.is_completed():
            assert task is None
            record.set_result(durable_promise.get_value(self._encoder), deduping=True)
            fork_or_join.handle.set_result(record.safe_result())
        else:
            return [Invoke(record.id), Subscribe(record.id, fork_or_join.handle)]

        return []

    def _handle_subscribe(self, subscribe: Subscribe) -> list[Command]:
        self._subsriptions.setdefault(subscribe.id, []).append(subscribe.handle)
        return []

    def _handle_notify(self, notify: Notify) -> list[Command]:
        for subscriber in self._subsriptions.pop(notify.id, []):
            subscriber.set_result(notify.value)
        return []

    def _handle_continue(
        self, id: str, next_value: Result[Any, Exception] | None
    ) -> list[Command]:
        record = self._records[id]
        coro = record.get_coro()
        yielded_value = coro.advance(next_value)

        if isinstance(yielded_value, LFI):
            return self._process_lfi(record, yielded_value)
        if isinstance(yielded_value, LFC):
            return self._process_lfc(record, yielded_value)
        if isinstance(yielded_value, RFI):
            return self._process_rfi(record, yielded_value)
        if isinstance(yielded_value, RFC):
            return self._process_rfc(record, yielded_value)
        if isinstance(yielded_value, Promise):
            return self._process_promise(record, yielded_value)
        if isinstance(yielded_value, FinalValue):
            return self._process_final_value(record, yielded_value.v)
        if isinstance(yielded_value, DI):
            # start execution from the top. Add current record to runnable
            raise NotImplementedError

        assert_never(yielded_value)

    def _process_invoke_msg(
        self, invoke_msg: InvokeMsg, task: TaskRecord
    ) -> list[Command]:
        logger.info(
            "Invoke message for %s received", invoke_msg.root_durable_promise.id
        )

        invoke_info = invoke_msg.root_durable_promise.invoke_info(self._encoder)
        func_name = invoke_info["func_name"]
        registered_func = self._registry.get(func_name)
        assert registered_func, f"There's no function registered under name {func_name}"
        func, opts = registered_func
        assert opts.durable

        rfi = RFI(
            Invocation(func, *invoke_info["args"], **invoke_info["kwargs"]), opts=opts
        )
        record = self._records.get(invoke_msg.root_durable_promise.id)
        if record:
            assert record.is_root
            assert isinstance(record.invocation, RFI)
            assert record.durable_promise is not None
            record.invocation = rfi
        else:
            record = Record[Any](
                id=invoke_msg.root_durable_promise.id,
                invocation=rfi,
                parent=None,
                ctx=Context(self._deps),
            )
            self._records[record.id] = record
            record.add_durable_promise(invoke_msg.root_durable_promise)

        record.add_task(task=task)
        return [Invoke(record.id)]

    def _process_resume_msg(
        self, resume_msg: ResumeMsg, task: TaskRecord
    ) -> list[Command]:
        logger.info(
            "Continue message for %s received", resume_msg.leaf_durable_promise.id
        )
        assert isinstance(self._store, RemoteStore)
        assert resume_msg.leaf_durable_promise.is_completed()

        root_record = self._records.get(resume_msg.root_durable_promise.id)
        leaf_record = self._records.get(resume_msg.leaf_durable_promise.id)
        if root_record and leaf_record:
            assert root_record.is_root
            assert leaf_record.is_root
            root_record.add_task(task)
            if not leaf_record.done():
                leaf_record.set_result(
                    resume_msg.leaf_durable_promise.get_value(self._encoder),
                    deduping=True,
                )

            return self._unblock_awaiting_remote(leaf_record.id)

        return self._process_invoke_msg(
            InvokeMsg(resume_msg.root_durable_promise), task
        )

    def _process_promise(
        self, record: Record[Any], promise: Promise[Any]
    ) -> list[Command]:
        loopback: list[Command] = []
        promise_record = self._records[promise.id]
        if promise_record.done():
            loopback.extend(
                self._handle_continue(record.id, promise_record.safe_result())
            )
        elif isinstance(promise_record.invocation, LFI):
            self._add_to_awaiting_local(promise_record.id, record.id)
        elif isinstance(promise_record.invocation, RFI):
            assert isinstance(self._store, RemoteStore)
            assert (
                promise_record.is_root
            ), "Callbacks can only be registered partition roots"

            root = record.root()
            leaf_id = promise_record.id

            assert self._recv
            assert (
                promise_record.durable_promise
            ), f"Record {promise_record.id} not backed by a promise"
            durable_promise, callback = self._store.callbacks.create(
                id=utils.string_to_uuid(record.id),
                promise_id=leaf_id,
                root_promise_id=root.id,
                timeout=sys.maxsize,
                recv=self._recv,
            )
            if callback is None:
                assert durable_promise.is_completed()
                record.set_result(
                    durable_promise.get_value(self._encoder), deduping=True
                )
            else:
                self._add_to_awaiting_remote(promise_record.id, record.id)
                if self._blocked_only_on_remote(root.id):
                    self._complete_task(root.id)

        else:
            assert_never(promise_record.invocation)

        return loopback

    def _unblock_awaiting_local(self, id: str) -> list[Command]:
        record = self._records[id]
        assert record.done()
        resume_cmds: list[Command] = []
        for blocked_id in self._awaiting_lfi.pop(record.id, []):
            blocked_record = self._records[blocked_id]
            assert blocked_record.blocked_on
            assert blocked_record.blocked_on.id == id
            blocked_record.blocked_on = None
            resume_cmds.append(Resume(blocked_id, record.safe_result()))
        return resume_cmds

    def _unblock_awaiting_remote(self, id: str) -> list[Command]:
        record = self._records[id]
        assert record.done()
        assert isinstance(record.invocation, RFI)
        resume_cmds: list[Command] = []
        for blocked_id in self._awaiting_rfi.pop(record.id, []):
            blocked_record = self._records[blocked_id]
            assert blocked_record.blocked_on
            assert blocked_record.blocked_on.id == id
            blocked_record.blocked_on = None
            resume_cmds.append(Resume(blocked_id, record.safe_result()))
        return resume_cmds

    def _add_to_awaiting_local(self, id: str, blocked: str) -> None:
        logger.info("Blocking %s awaiting locally on %s", blocked, id)
        record = self._records[id]
        assert isinstance(record.invocation, LFI)
        assert not record.done()
        blocked_record = self._records[blocked]
        assert not blocked_record.blocked_on
        blocked_record.blocked_on = record
        self._awaiting_lfi.setdefault(id, []).append(blocked)

    def _add_to_awaiting_remote(self, id: str, blocked: str) -> None:
        logger.info("Blocking %s awaiting remotely on %s", blocked, id)
        record = self._records[id]
        assert isinstance(record.invocation, RFI)
        assert not record.done()
        blocked_record = self._records[blocked]
        assert not blocked_record.blocked_on
        blocked_record.blocked_on = record
        self._awaiting_rfi.setdefault(id, []).append(blocked)

    def _blocked_only_on_remote(self, id: str, *, root: bool = True) -> bool:
        record = self._records[id]
        if isinstance(record.invocation, RFI) and not root:
            return True
        if not record.blocked_on:
            return False
        return all(
            self._blocked_only_on_remote(child.id, root=False)
            for child in record.children
            if not child.done()
        )

    def _complete_task(self, id: str) -> None:
        record = self._records[id]
        assert record.has_task()

        assert isinstance(self._store, RemoteStore)
        task = record.get_task()
        self._store.tasks.complete(
            task_id=task.task_id,
            counter=task.counter,
        )
        record.remove_task()

    def _process_rfc(self, record: Record[Any], rfc: RFC) -> list[Command]:  # noqa: PLR0912
        loopback: list[Command] = []
        assert isinstance(self._store, RemoteStore)
        child_id: str
        next_child_name = record.next_child_name()
        if isinstance(rfc.unit, Invocation):
            child_id = rfc.opts.id if rfc.opts.id is not None else next_child_name
        elif isinstance(rfc.unit, DurablePromise):
            child_id = rfc.unit.id if rfc.unit.id is not None else next_child_name
        else:
            assert_never(rfc.unit)
        child_record = self._records.get(child_id)
        root = record.root()
        if child_record is not None:
            record.add_child(child_record)
            if child_record.done():
                loopback.extend(
                    self._handle_continue(record.id, child_record.safe_result())
                )
            else:
                self._add_to_awaiting_remote(child_id, record.id)
                if self._blocked_only_on_remote(root.id):
                    self._complete_task(root.id)
        else:
            child_record = record.create_child(id=child_id, invocation=rfc.to_rfi())
            self._records[child_id] = child_record
            assert rfc.opts.durable

            data, headers, tags, timeout = self._get_info_from_rfi(rfc.to_rfi())

            assert self._recv

            durable_promise = self._store.promises.create(
                id=child_id,
                ikey=utils.string_to_uuid(child_id),
                strict=False,
                headers=headers,
                data=data,
                timeout=timeout or sys.maxsize,
                tags=tags,
            )
            assert child_id in self._records
            assert not record.done()
            child_record.add_durable_promise(durable_promise)

            if durable_promise.is_completed():
                value = durable_promise.get_value(self._encoder)
                child_record.set_result(value, deduping=True)
                loopback.extend(
                    self._handle_continue(record.id, child_record.safe_result())
                )

            else:
                durable_promise, callback = self._store.callbacks.create(
                    id=utils.string_to_uuid(record.id),
                    promise_id=child_id,
                    root_promise_id=root.id,
                    timeout=sys.maxsize,
                    recv=self._recv,
                )
                if durable_promise.is_completed():
                    assert callback is None
                    value = durable_promise.get_value(self._encoder)
                    child_record.set_result(value, deduping=True)
                    loopback.extend(
                        self._handle_continue(record.id, child_record.safe_result())
                    )
                else:
                    assert callback is not None
                    self._add_to_awaiting_remote(child_id, record.id)
                    if self._blocked_only_on_remote(root.id):
                        self._complete_task(root.id)

        return loopback

    def _process_lfc(self, record: Record[Any], lfc: LFC) -> list[Command]:
        loopback: list[Command] = []
        child_id = lfc.opts.id if lfc.opts.id is not None else record.next_child_name()
        child_record = self._records.get(child_id)
        if child_record is not None:
            record.add_child(child_record)
            if child_record.done():
                loopback.extend(
                    self._handle_continue(record.id, child_record.safe_result())
                )
            else:
                self._add_to_awaiting_local(child_id, record.id)

        else:
            child_record = record.create_child(id=child_id, invocation=lfc.to_lfi())
            self._records[child_id] = child_record

            if lfc.opts.durable:
                durable_promise = self._store.promises.create(
                    id=child_id,
                    ikey=utils.string_to_uuid(child_id),
                    strict=False,
                    headers=None,
                    data=None,
                    timeout=sys.maxsize,
                    tags=None,
                )
                assert child_id in self._records
                assert not record.done()
                child_record.add_durable_promise(durable_promise)
                if durable_promise.is_completed():
                    value = durable_promise.get_value(self._encoder)
                    child_record.set_result(value, deduping=True)
                    loopback.extend(
                        self._handle_continue(record.id, child_record.safe_result())
                    )
                else:
                    loopback.append(Invoke(child_id))
                    self._add_to_awaiting_local(child_id, record.id)

            else:
                loopback.append(Invoke(child_id))
                self._add_to_awaiting_local(child_id, record.id)
        return loopback

    def _process_rfi(self, record: Record[Any], rfi: RFI) -> list[Command]:
        loopback: list[Command] = []
        child_id: str
        next_child_name = record.next_child_name()
        if isinstance(rfi.unit, Invocation):
            child_id = rfi.opts.id if rfi.opts.id is not None else next_child_name
        elif isinstance(rfi.unit, DurablePromise):
            child_id = rfi.unit.id if rfi.unit.id is not None else next_child_name
        else:
            assert_never(rfi.unit)
        child_record = self._records.get(child_id)
        if child_record is not None:
            record.add_child(child_record)
            loopback.extend(
                self._handle_continue(record.id, Ok(Promise[Any](child_record.id)))
            )
        else:
            child_record = record.create_child(id=child_id, invocation=rfi)
            self._records[child_id] = child_record
            assert rfi.opts.durable

            data, headers, tags, timeout = self._get_info_from_rfi(rfi)

            durable_promise = self._store.promises.create(
                id=child_id,
                ikey=utils.string_to_uuid(child_id),
                strict=False,
                headers=headers,
                data=data,
                timeout=timeout or sys.maxsize,
                tags=tags,
            )
            assert child_id in self._records
            assert not record.done()
            child_record.add_durable_promise(durable_promise)

            if durable_promise.is_completed():
                value = durable_promise.get_value(self._encoder)
                child_record.set_result(value, deduping=True)
            loopback.extend(
                self._handle_continue(record.id, Ok(Promise[Any](child_record.id)))
            )

        return loopback

    def _process_lfi(self, record: Record[Any], lfi: LFI) -> list[Command]:
        loopback: list[Command] = []
        child_id = lfi.opts.id if lfi.opts.id is not None else record.next_child_name()
        child_record = self._records.get(child_id)
        if child_record is not None:
            record.add_child(child_record)
            loopback.extend(
                self._handle_continue(record.id, Ok(Promise[Any](child_record.id)))
            )
        else:
            child_record = record.create_child(id=child_id, invocation=lfi)
            self._records[child_id] = child_record

            if lfi.opts.durable:
                durable_promise = self._store.promises.create(
                    id=child_id,
                    ikey=utils.string_to_uuid(child_id),
                    strict=False,
                    headers=None,
                    data=None,
                    timeout=sys.maxsize,
                    tags=None,
                )
                assert child_id in self._records
                assert not record.done()

                child_record.add_durable_promise(durable_promise)

                if durable_promise.is_completed():
                    value = durable_promise.get_value(self._encoder)
                    child_record.set_result(value, deduping=True)
                else:
                    loopback.append(Invoke(child_id))
                loopback.extend(
                    self._handle_continue(record.id, Ok(Promise[Any](child_record.id)))
                )
            else:
                loopback.append(Invoke(child_id))
                loopback.extend(
                    self._handle_continue(record.id, Ok(Promise[Any](child_record.id)))
                )

        return loopback

    def _process_final_value(
        self, record: Record[Any], final_value: Result[Any, Exception]
    ) -> list[Command]:
        return [Complete(record.id, final_value)]

    def _get_info_from_rfi(self, rfi: RFI) -> tuple[Data, Headers, Tags, int | None]:
        data: Data
        tags: Tags
        headers: Headers
        timeout: int | None = None
        if isinstance(rfi.unit, DurablePromise):
            data = rfi.unit.data
            tags = rfi.unit.tags
            headers = rfi.unit.headers
            timeout = rfi.unit.timeout
        elif isinstance(rfi.unit, Invocation):
            func: str
            if isinstance(rfi.unit.fn, str):
                func = rfi.unit.fn
            else:
                assert isfunction(rfi.unit.fn)
                registered_fn_name = self._registry.get_from_value(rfi.unit.fn)
                assert registered_fn_name is not None
                func = registered_fn_name
            data = self._encoder.encode(
                {
                    "func": func,
                    "args": rfi.unit.args,
                    "kwargs": rfi.unit.kwargs,
                }
            )
            tags = {"resonate:invoke": rfi.opts.send_to or "default"}
            headers = None
        else:
            assert_never(rfi.unit)
        return (data, headers, tags, timeout)

    def _stop(self) -> None:
        self._cmd_queue.put(None)
        self._scheduler_thread.join()

    def stop(self) -> None:
        self._delay_queue.stop()
        self._stop()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from resonate.stores.record import (
        DurablePromiseRecord,
        TaskRecord,
    )
    from resonate.typing import Data, Headers, IdempotencyKey, Tags


class IPromiseStore(ABC):
    @abstractmethod
    def create(  # noqa: PLR0913
        self,
        *,
        id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
    ) -> DurablePromiseRecord: ...
    @abstractmethod
    def create_with_task(  # noqa: PLR0913
        self,
        *,
        id: str,
        ikey: str | None,
        strict: bool,
        headers: Headers,
        data: Data,
        timeout: int,
        tags: Tags,
        pid: str,
        ttl: int,
    ) -> tuple[DurablePromiseRecord, TaskRecord | None]: ...

    @abstractmethod
    def resolve(
        self,
        *,
        id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord: ...

    @abstractmethod
    def reject(
        self,
        *,
        id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord: ...

    @abstractmethod
    def cancel(
        self,
        *,
        id: str,
        ikey: IdempotencyKey,
        strict: bool,
        headers: Headers,
        data: Data,
    ) -> DurablePromiseRecord: ...

    @abstractmethod
    def get(self, *, id: str) -> DurablePromiseRecord: ...



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union, final

from typing_extensions import TypeAlias


class Retriable(ABC):
    @abstractmethod
    def calculate_delay(self, attempt: int) -> float: ...

    @abstractmethod
    def should_retry(self, attempt: int) -> bool: ...


@final
@dataclass(frozen=True)
class Exponential(Retriable):
    """A retry policy where the delay between retries grows exponentially."""

    base_delay: float
    factor: float
    max_delay: float
    max_retries: int

    def calculate_delay(self, attempt: int) -> float:
        assert attempt > 0, "Attempt must be positive."
        return min(self.base_delay * (self.factor**attempt), self.max_delay)

    def should_retry(self, attempt: int) -> bool:
        assert attempt > 0, "Attempt must be positive."
        if self.max_retries < 0:
            return True
        return attempt <= self.max_retries


@final
@dataclass(frozen=True)
class Linear(Retriable):
    """A retry policy where the delay between retries grows linearly."""

    delay: float
    max_retries: int

    def calculate_delay(self, attempt: int) -> float:
        assert attempt > 0, "Attempt must be positive."
        return self.delay * attempt

    def should_retry(self, attempt: int) -> bool:
        assert attempt > 0, "Attempt must be positive."
        if self.max_retries < 0:
            return True
        return attempt <= self.max_retries


@final
@dataclass(frozen=True)
class Constant(Retriable):
    """A retry policy where the delay between retries is constant."""

    delay: float
    max_retries: int

    def calculate_delay(self, attempt: int) -> float:
        assert attempt > 0, "Attempt must be positive."
        return self.delay

    def should_retry(self, attempt: int) -> bool:
        assert attempt > 0, "Attempt must be positive."
        if self.max_retries < 0:
            return True
        return attempt <= self.max_retries


@final
@dataclass(frozen=True)
class Never:
    """A retry policy where there's no retry."""


RetryPolicy: TypeAlias = Union[Exponential, Linear, Constant, Never]


def linear(delay: float, max_retries: int) -> Linear:
    return Linear(delay=delay, max_retries=max_retries)


def exponential(
    base_delay: float, factor: float, max_delay: float, max_retries: int
) -> Exponential:
    return Exponential(
        base_delay=base_delay,
        factor=factor,
        max_retries=max_retries,
        max_delay=max_delay,
    )


def constant(delay: float, max_retries: int) -> Constant:
    return Constant(delay=delay, max_retries=max_retries)


def never() -> Never:
    return Never()



"""Execution events."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Union, final

from typing_extensions import ParamSpec, TypeAlias

if TYPE_CHECKING:
    from resonate.result import Result

P = ParamSpec("P")


# Promise Events


@final
@dataclass(frozen=True)
class PromiseCreated:
    id: str
    parent_id: str | None
    tick: int


@final
@dataclass(frozen=True)
class PromiseCompleted:
    id: str
    parent_id: str | None
    tick: int
    value: Result[Any, Exception]


# Execution Events


@final
@dataclass(frozen=True)
class ExecutionInvoked:
    id: str
    parent_id: str | None
    tick: int
    fn_name: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]


@final
@dataclass(frozen=True)
class ExecutionTerminated:
    id: str
    parent_id: str | None
    tick: int


@final
@dataclass(frozen=True)
class ExecutionResumed:
    id: str
    parent_id: str | None
    tick: int


@final
@dataclass(frozen=True)
class ExecutionAwaited:
    id: str
    parent_id: str | None
    tick: int


SchedulerEvents: TypeAlias = Union[
    PromiseCreated,
    PromiseCompleted,
    ExecutionInvoked,
    ExecutionTerminated,
    ExecutionResumed,
    ExecutionAwaited,
]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, final

if TYPE_CHECKING:
    from resonate import retry_policy


@final
@dataclass
class Options:
    id: str | None = None
    durable: bool = True
    send_to: str | None = None
    retry_policy: retry_policy.RetryPolicy | None = None
    version: int = 1
    execute_here: bool = True



from __future__ import annotations

from resonate import random


def test_random_prefix() -> None:
    choices_per_run = 10
    global_prefix: list[float] = []
    steps_taken: list[int] = []
    for i in range(30):
        rand = random.Random(seed=i, prefix=global_prefix)
        remembered_steps = len(global_prefix)
        steps_taken_this_run = [
            rand.randint(0, 100) for _ in range(remembered_steps + choices_per_run)
        ]
        global_prefix.extend(rand.export()[remembered_steps:])
        assert (
            len(steps_taken_this_run) - len(global_prefix) == 0
        ), f"Every run must make {choices_per_run} new choices"

        steps_taken.extend(steps_taken_this_run[remembered_steps:])
        assert (
            steps_taken[:remembered_steps] == steps_taken_this_run[:remembered_steps]
        ), f"The first {remembered_steps} must always be equal."


def test_random_prefix_with_floats() -> None:
    choices_per_run = 10
    global_prefix: list[float] = []
    steps_taken: list[float] = []
    for i in range(30):
        rand = random.Random(seed=i, prefix=global_prefix)
        remembered_steps = len(global_prefix)
        steps_taken_this_run = [
            rand.random() for _ in range(remembered_steps + choices_per_run)
        ]
        global_prefix.extend(rand.export()[remembered_steps:])
        assert (
            len(steps_taken_this_run) - len(global_prefix) == 0
        ), f"Every run must make {choices_per_run} new choices"

        steps_taken.extend(steps_taken_this_run[remembered_steps:])
        assert (
            steps_taken[:remembered_steps] == steps_taken_this_run[:remembered_steps]
        ), f"The first {remembered_steps} must always be equal."



from __future__ import annotations

from resonate.context import Context
from resonate.dataclasses import DurablePromise
from resonate.handle import Handle
from resonate.promise import Promise
from resonate.resonate import Resonate

__all__ = ["Resonate", "Handle", "DurablePromise", "Promise", "Context"]



from __future__ import annotations

from resonate.retry_policy import (
    constant,
    exponential,
    linear,
)


def test_infinite_retry() -> None:
    exponential_policy = exponential(
        base_delay=1, factor=2, max_retries=-1, max_delay=30
    )
    constant_policy = constant(delay=1, max_retries=-1)
    linear_policy = linear(delay=1, max_retries=-1)
    for i in range(10, 1000, 10):
        assert exponential_policy.should_retry(i)
        assert constant_policy.should_retry(i)
        assert linear_policy.should_retry(i)


def test_exponential() -> None:
    policy = exponential(base_delay=1, factor=2, max_retries=5, max_delay=8)
    expected_delays = (2, 4, 8, 8, 8)
    assert policy.should_retry(attempt=1)
    assert policy.calculate_delay(attempt=1) == expected_delays[0]
    assert policy.should_retry(attempt=2)
    assert policy.calculate_delay(attempt=2) == expected_delays[1]
    assert policy.should_retry(attempt=3)
    assert policy.calculate_delay(attempt=3) == expected_delays[2]
    assert policy.should_retry(attempt=4)
    assert policy.calculate_delay(attempt=4) == expected_delays[3]
    assert policy.should_retry(attempt=5)
    assert policy.calculate_delay(attempt=5) == expected_delays[4]
    assert not policy.should_retry(attempt=6)


def test_linear() -> None:
    policy = linear(delay=2, max_retries=3)
    expected_delays = (2, 4, 6)
    assert policy.should_retry(attempt=1)
    assert policy.calculate_delay(attempt=1) == expected_delays[0]
    assert policy.should_retry(attempt=2)
    assert policy.calculate_delay(attempt=2) == expected_delays[1]
    assert policy.should_retry(attempt=3)
    assert policy.calculate_delay(attempt=3) == expected_delays[2]
    assert not policy.should_retry(attempt=4)


def test_constant() -> None:
    policy = constant(delay=2, max_retries=3)
    expected_delays = (2, 2, 2)
    assert policy.should_retry(attempt=1)
    assert policy.calculate_delay(attempt=1) == expected_delays[0]
    assert policy.should_retry(attempt=2)
    assert policy.calculate_delay(attempt=2) == expected_delays[1]
    assert policy.should_retry(attempt=3)
    assert policy.calculate_delay(attempt=3) == expected_delays[2]
    assert not policy.should_retry(attempt=4)
